/*
 * buntichboot.c
 *
 * Created: 08.11.2011 23:57:56
 *  Author: tut
 *
 * BOOTSZ1 BOOTSZ0  Boot Size
 *    1       0     256 words $0000-$1EFF $1F00-$1FFF
 *>   0       1     512 words $0000-$1DFF $1E00-$1FFF < this is active
 *
 *
 * Fuse-Setting for Buntich:
 * LFUSE = 0x3f
 * HFUSE = 0xc2
 * LOCK  = 0xef     Avoid writing to bootloader
 */ 
#ifndef F_CPU
    #define F_CPU 16000000UL
#endif

#define BOOTLOADER_ADDRESS (0x3c00U)

#include <stdint.h>
#include <avr/io.h>
#include <avr/boot.h>
 
#define PINOFF(PORT,NR) PORT&=~(1 << NR)
#define PINON(PORT,NR)  PORT|=(1 << NR)

#define BUNTICH_CHIP_ENUMIN_PORT PORTD
#define BUNTICH_CHIP_ENUMIN_DDR DDRD
#define BUNTICH_CHIP_ENUMIN_PIN PIND
#define BUNTICH_CHIP_ENUMIN_NUMBER 4

#define BUNTICH_CHIP_ENUMOUT_PORT PORTD
#define BUNTICH_CHIP_ENUMOUT_DDR DDRD
#define BUNTICH_CHIP_ENUMOUT_PIN PIND
#define BUNTICH_CHIP_ENUMOUT_NUMBER 5

#define BUNTICH_CHIP_TERM_PORT PORTD
#define BUNTICH_CHIP_TERM_DDR DDRD
#define BUNTICH_CHIP_TERM_PIN PIND
#define BUNTICH_CHIP_TERM_NUMBER 7

#define BUNTICH_CHIP_TXENABLE_PORT PORTD
#define BUNTICH_CHIP_TXENABLE_DDR DDRD
#define BUNTICH_CHIP_TXENABLE_PIN PIND
#define BUNTICH_CHIP_TXENABLE_NUMBER 2

uint8_t rxbuf[520];

// Timeout stuff
    #define TimeoutReached()	(TIFR & (1<<OCF1A))

    // Use this to define Timeouts up to 4s in 64us steps
    #define SetTimeout_us(T)		SetNextTimeout((T)/((1024UL * 1000000UL) / F_CPU ))

    void SetNextTimeout(uint16_t ticks)
    {
        TCNT1 = 0;
        OCR1A = ticks;
        PINON(TIFR,OCF1A);
    }

    #define IsUARTByteAvailible() (UCSRA & (1<<RXC)) 
    
// UART
    uint8_t GetUARTByte(void)
    {
        /* Wait for data to be received */
        while ( !(UCSRA & (1<<RXC)) )
        ;

        /* Get and return received data from buffer */
        return UDR;
    }

    void SendUARTByte(uint8_t msg)
    {
        while ( !( UCSRA & (1<<UDRE)) )
            ;
        PINON(BUNTICH_CHIP_TXENABLE_PORT,BUNTICH_CHIP_TXENABLE_NUMBER);	// enable
        PINON(UCSRA,TXC);
        UDR=msg;
        while ( !( UCSRA & (1<<TXC)) )
            ;
        PINOFF(BUNTICH_CHIP_TXENABLE_PORT,BUNTICH_CHIP_TXENABLE_NUMBER);	// disable
    }

// Other stuff	
    // Waits until enum-in gets low and activate enum-out short time afterwards
    void TxByte(uint8_t msg)
    {
        SetTimeout_us(100000); 	// 100ms max. wait time
        while(!TimeoutReached() && (BUNTICH_CHIP_ENUMIN_PIN & (1<<BUNTICH_CHIP_ENUMIN_NUMBER))) ;
        if(TimeoutReached()) return;
        SendUARTByte(msg);
        PINOFF(BUNTICH_CHIP_ENUMOUT_PORT, BUNTICH_CHIP_ENUMOUT_NUMBER);
        SetTimeout_us(5000);
        while(!TimeoutReached()) ;
        PINON(BUNTICH_CHIP_ENUMOUT_PORT, BUNTICH_CHIP_ENUMOUT_NUMBER);	
    }

    void ExitBootloader(void)
    {
     //   void (*start)( void ) = BOOTLOADER_ADDRESS;        
        void (*start)( void ) = 0x0000;        /* Function pointer to 0x0000 */
        asm volatile("cli;");
        UCSRB = 0;
        PORTC = 0xf0;
        boot_rww_enable();
        SetTimeout_us(100000);
        while(!TimeoutReached()) ;
     //   while(1) ;
        start();
    }

    // Write 4 pages = 512 byte from buffer 
    void Write4Pages(uint16_t address)
    {
        uint8_t i,j;
        uint8_t *buf = rxbuf;
        address &= 0xff80;
        
        for(j=0;j<4;j++)
        {
            eeprom_busy_wait ();

            boot_page_erase (address);
            boot_spm_busy_wait ();      // Wait until the memory is erased.

            for (i=0; i<SPM_PAGESIZE; i+=2)
            {
                // Set up little-endian word.

                uint16_t w = *buf++;
                w += (*buf++) << 8;
            
                boot_page_fill (address + i, w);
            }

            boot_page_write (address);     // Store buffer in flash page.
            boot_spm_busy_wait();       // Wait until the memory is written.
            address += SPM_PAGESIZE;
        }
    }
    
int main()
{
    uint8_t c, checksum;
    uint16_t blockaddress,i;

    asm volatile("cli;");
    
     // LED Ports
    DDRA = 0xff; // 1=output
    DDRB = 0xff;
    DDRC = 0xff;
    PORTA = 0xff; // 1 while input will activate pull-up, 1 = LED OFF
    PORTB = 0xff;
    PORTC = 0xff;
    
    // Other ports:
    // 7 IP TERM (Jumper)
    // 6 O1 ENUM2D
    // 5 O1 ENUMOUT
    // 4 IP ENUMIN
    // 3 O0 COMMONCONTROL (0 = LEDs activated)
    // 2 O0 TXENABLE (1 = enabled)
    // 1 O1 TX
    // 0 IP RX
    DDRD  = 0b01101110;
    PORTD = 0b11110011;
    
    // UART Init
    UCSRA = 0x20;	// normal speed, no interrupt
    UCSRB = 0x18;	// Rx/Tx, 8 bit
    UCSRC = 0x86;	// 8 bit, no parity
    UBRRH = 0;	// 1Mbps
    UBRRL = 0;
    
    // Timer1 init 
    TCCR1A = 0;
    TCCR1B = 5; // Clk/1024
    TIMSK = 0;
    
    PORTC = 0x7f;
    SetTimeout_us(100000);
    while(!TimeoutReached()) ;
    TxByte('R');
    
    // Init sequenz:
    // - leave after 16 rx byte or 4 seconds
    // - enter BL only if last four byte == code [0xEE 0xDD 0x11 0xAA]
    //   and no rxbyte for 100ms
    {
        uint8_t rxcount = 0, exitflag = 0;
        SetTimeout_us(4000000UL);
        while(!exitflag)
        {
            if(TimeoutReached()) ExitBootloader();
            if(IsUARTByteAvailible())
            {
                rxbuf[rxcount++] = GetUARTByte();
                if(rxcount > 16) ExitBootloader();
                if(rxcount >= 4) 
                {
                    if((rxbuf[rxcount-4] == 0xEE) && 
                       (rxbuf[rxcount-3] == 0xDD) && 
                       (rxbuf[rxcount-2] == 0x11) && 
                       (rxbuf[rxcount-1] == 0xAA)) exitflag = 1;
                }
            }
        } 
        PORTC = 0xbf;
        SetTimeout_us(100000UL);
        while(!TimeoutReached())
        {
            if(IsUARTByteAvailible()) ExitBootloader();
        }
    }
    
    // Bootloader entered, send reply (can be used as BL version indication)
    TxByte('1');
    PORTC = 0xdf;
    
    for(;;)
    {
        c = GetUARTByte();
        switch((unsigned char)c)
        {
            case 'p': // Program 4 pages 
                SetTimeout_us(20000UL);
                checksum = blockaddress = GetUARTByte();
                c = GetUARTByte();
                checksum += c;
                blockaddress |= ((uint16_t)c << 8);
                for( i = 0;i<512;i++)
                {
                    while(!IsUARTByteAvailible() && !TimeoutReached())
                    {
                    }
                    if(TimeoutReached()) break;
                    c = GetUARTByte();
                    checksum += c;  
                    rxbuf[i] = c;       
                }
                c = GetUARTByte();
                if((c == checksum) && (i==512) && (blockaddress <= (BOOTLOADER_ADDRESS-512)))
                {
                    Write4Pages(blockaddress);
                    TxByte('o');
                }
                else
                {
                    TxByte('f');
                }
                break;
                
            case 'r': // Do reset to main program
                TxByte('o');
                ExitBootloader();
                break;
               
            case 'n': // No operation, returns o
                TxByte('o');
                break;
               
            default:
                break;
        }
    }

    return 0;
}



